#! /usr/bin/python2

# $Id: rpmWrapper.py,v 1.13 2003/07/28 21:19:02 ensc Exp $
# Copyright (C) 2001 Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
#  
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#  
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#  
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#  

import rpm
import re
import os
import string
import sys

from tools import unique
import config
import params

RE_RPM_FULLNAMES = [ re.compile('^(.+)-([^-]+-[^-]+)'),
#                     re.compile('^(.+)-(\d[^-]*)'),
                     re.compile('^(.+)') ]

ambiguous_deps = []

_isDirectory = (lambda mode:(mode&0170000)==040000 or (mode&0120000)==0120000)
"""Tells if a file with <mode> is a directory."""

def _splitPath(path):
    """Returns <path> as a (basedir, basename) pair."""
    
    pos = string.rfind(path,'/')
    if pos>0:    basedir=path[:pos]
    elif pos==0: basedir='/'
    else:        basedir=''

    if pos+1<len(path): basename=path[pos+1:]
    else:               basename=''

    return (basedir, basename)

def _buildName(hdr):
    return hdr[rpm.RPMTAG_NAME] + '-' + hdr.dsOfHeader().EVR()
    
class file:
    """Encapsulates information about a file handled by rpm. Its
    methods are expecting output from 'rpm -q --dump'."""
    
    def __init__(this, name):
        """Constructs the object which identifies the file given by
        str."""

        (this.__basedir, this.__basename) =  _splitPath(name)
        this.__name = name

    def getName(this):
        """Tells the filename of the encapsulated file."""
        
        return this.__name

    def getBasename(this):
        """Tells the plain filename (basename) of the encapsulated
        file."""
        
        return this.__basename
    
    def getPath(this):
        """Tells the path without the basename of the encapsulated
        file."""
        
        return this.__basedir

_deps = {}
_ts   = None

def getDB():
    """Singleton method; returns the opened rpm-db and sets various
    options (e.g. _dbpath) when opening it the first time."""
    
    global _ts
    
    if _ts==None:
        if config.cfg.dbpath: rpm.addMacro("_dbpath", config.cfg.dbpath)
        _ts = rpm.TransactionSet()
        _ts.setVSFlags(-1)

    return _ts

def getAllPackages():
    """Returns a list of all packages in the current rpm-db. The
    packages are represented by their indices in the rpm-db."""

    result = []
    for hdr in getDB().dbMatch():
        if hdr[rpm.RPMTAG_NAME] in params.IGNORES: continue
        
        result.append(_buildName(hdr))

    return result

class package:
    """Encapsulates information about a rpm-package and provides access
    to them. Objects will be created by the name of the rpm-package or
    their indices in the rpm-db.

    The name,version and release of a package are available through a
    proxy-object returned by getData(); information about contained
    files by the getFiles() method.

    TODO: currently only the canonical form <name>-<version>-<release>
          or the simple <name> can be used as the name of an rpm. The
          <name>-<version> format is not possible."""

    class __dataProxy:
        """Encapsulates general information about a rpm-package. Attributes
        are name, serial, version and release."""

        def __init__(this, name, serial, version, release):
            this.name     = name
            this.serial   = serial
            this.version  = version
            this.release  = release
            this.fullname = '%s-%s-%s' % (name, version, release)

    class __fileProxy:
        """Encapsulates information about the files of an
        rpm-package. Available methods are getDirectories() returning
        a list of directories owned by the package, and getPaths()
        returning a list of all paths used by the package."""
        
        def __init__(this, header):
            """Constructs a __fileProxy object providing access to
            information in the rpmdb header."""

            this.__directories = this.__getDirectories(header)
            this.__paths       = this.__getPaths(header)

        getDirectories = lambda this: this.__directories
        """Returns a list of directories owned by the package."""

        getPaths = lambda this: this.__paths
        """Returns a list of all paths used by the package, including
        all subpaths. The result contains each path exactly once."""


        def __getDirectories(this,header):
            names = header[rpm.RPMTAG_FILENAMES]
            modes = header[rpm.RPMTAG_FILEMODES]

            result = []
            for i in xrange(0, len(names)):
                if _isDirectory(modes[i]):
                    result.append( file(names[i]) )

            return result

        def __getSubpaths(this, path):
            """Returns a list of all paths where <path> is located
            in. E.g. it returns ['/', '/usr', '/usr/share'] for
            '/usr/share/doc'."""
            
            pos = 0
            result = []
            while 1:
                pos = string.find(path, '/', pos+1)
                if pos==-1: break

                result.append(path[:pos])

            return result

        def __getPaths(this,header):
            names  = header[rpm.RPMTAG_FILENAMES]
            
            if names: result = ['/']
            else:     result = []

            for name in names:
                result.append(_splitPath(name)[0])

            result = unique(result)
            for i in result:
                result.extend(this.__getSubpaths(i))
                
            return unique(result)
        
    def __init__(this, id):
        """Construct an rpm-package object identified by <id>. If <id>
        is a number it is used as an index in the rpmdb. Else
        __findIdxByFullName() will be used to find it in the
        rpm-db.

        Then this constructor tries to fetch the corresponding dataset
        from the rpm-db."""
        
        if type(id)==type(""): header = this.__findHeaderByFullName(id)
        else:                  header = id

        this.__deps   = this.__calculateDependencies(header)
        this.__files  = this.__fileProxy(header)
        this.__data   = this.__dataProxy(header[rpm.RPMTAG_NAME],
                                         header[rpm.RPMTAG_SERIAL],
                                         header[rpm.RPMTAG_VERSION],
                                         header[rpm.RPMTAG_RELEASE])
        this.__id     = _buildName(header)


    def __calculateDependencies(this,header):
        deps      = []
        plaindeps = []
        ds        = header.dsFromHeader()

        for i in xrange(0,ds.Count()):
            plaindeps.append(ds.next())

        if plaindeps!=None:
            for (name,evr,flags) in plaindeps:
                if (flags & rpm.RPMSENSE_RPMLIB)==rpm.RPMSENSE_RPMLIB:
                    attr=rpm.RPMTAG_PROVIDENAME
                    name="rpm"
                elif name!='' and name[0]=='/': attr=rpm.RPMTAG_BASENAMES
                else:                           attr=rpm.RPMTAG_PROVIDENAME
                
                deps.extend(this.__getDependenciesCached(header,attr,name))

        return unique(deps)
        

    def __splitFullName(this, name):
        """Called by __findHeaderByFullName() to split the name into a
        (name, version, release) tuple. If this is not possible,
        `None' will be returned."""
        
        for regexp in RE_RPM_FULLNAMES:
            m = regexp.match(name)
            if m:
                tmp = m.groups() + ("", "", "")
                return (tmp[0], tmp[1])
    
        return None

    def __findHeaderByFullName(this, name):
        """Searches for the rpm-package with name <name> in the
        rpm-db. Valid formats are <name>-<version>-<release> and
        <name>. <name>-<version> is NOT possible.

        If exactly one package was found, its index in the rpm-db will
        be returned. Else an exception will be thrown.

        It calls __splitFullName to split the <name>."""

        tmp = this.__splitFullName(name)
        if tmp==None: raise "Invalid packagename '%s'" % name
        ds  = rpm.dsSingle(rpm.RPMTAG_PROVIDENAME, tmp[0], tmp[1])
        evr = ds.EVR()

        pkgs = []

        for hdr in getDB().dbMatch(rpm.RPMTAG_NAME, tmp[0]):
            if not evr or hdr.dsOfHeader().EVR()==evr:  pkgs.append(hdr)

        if   len(pkgs)==0: raise "No such package '%s'" % name
        elif len(pkgs) >1: raise "Multiple packages are matching string '%s'" % name

        return pkgs[0]

    def __getDependenciesCached(this, header, attr, name):
        global _deps;
        depnames = _deps.get( (attr,name) )
        if depnames==None:
            depnames = {}
            for hdr in getDB().dbMatch(attr, name):
                depnames[ hdr[rpm.RPMTAG_NAME] ] = hdr

            _deps[ (attr,name) ] = depnames

        aux      = depnames.copy()
        selfname = header[rpm.RPMTAG_NAME]
        
        for (sup,sub) in params.RPM_PREFS:
            if aux.has_key(sup) and aux.has_key(sub):
                if selfname==sub: del aux[sup]
                else:             del aux[sub]

        res = map(lambda x: _buildName(x), aux.values())
        if len(res)>1: ambiguous_deps.append(res)

        #print "[%s,%s] -> %s" % (attr,name,res)
        return res


    getId = lambda this: this.__id

    getDependencies = lambda this: this.__deps
    """Returns packages required by the encapsulated package as a list
    of package-names."""


    getFiles = lambda this: this.__files
    """Returns the files contained in the encapsulated package as a
    __fileProxy object. There the attributes `directories' and `paths'
    are describing the owned directories and the used paths of the
    package."""
    
    getData = lambda this: this.__data
    """Returns a __dataProxy object encapsulating general information
    about the rpm-package."""

    __str__ = lambda this: this.__data.fullname


### Local Variables:
### mode: python
### End:
